import os
from shutil import copy2
import pytest

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
RASTERGIS_DATA_DIR = os.path.join(DATA_DIR, "rastergis")
CLASSIFICATION_DATA_DIR = os.path.join(DATA_DIR, "classification")
IMGCALC_DATA_DIR = os.path.join(DATA_DIR, "imagecalc")


def test_get_rat_length():
    import rsgislib.rastergis

    ref_clumps_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_attref.kea"
    )

    n_clumps = rsgislib.rastergis.get_rat_length(ref_clumps_img)
    assert n_clumps == 11949


def test_get_rat_columns():
    import rsgislib.rastergis

    ref_clumps_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_attref.kea"
    )

    rat_columns = rsgislib.rastergis.get_rat_columns(ref_clumps_img)
    ref_columns = ["Histogram", "Red", "Green", "Blue", "Alpha"]
    for i in [1, 2, 3, 4]:
        ref_columns.append("b{}Min".format(i))
        ref_columns.append("b{}Max".format(i))
        ref_columns.append("b{}Mean".format(i))
        ref_columns.append("b{}Sum".format(i))
        ref_columns.append("b{}StdDev".format(i))

    cols_match = True
    if len(rat_columns) != len(ref_columns):
        cols_match = False
        print(
            "Number of columns is different: {} != {}".format(
                len(rat_columns), len(ref_columns)
            )
        )

    if cols_match:
        for col in rat_columns:
            if col not in ref_columns:
                print("Col '{}' is not within the reference list".format(col))
                cols_match = False
                break
    assert cols_match


def test_get_rat_columns_info():
    import rsgislib.rastergis
    from osgeo import gdal

    ref_clumps_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_attref.kea"
    )

    rat_columns_info = rsgislib.rastergis.get_rat_columns_info(ref_clumps_img)

    correct_info = True
    if rat_columns_info["Histogram"]["type"] != gdal.GFT_Real:
        correct_info = False
    if rat_columns_info["Histogram"]["usage"] != gdal.GFU_PixelCount:
        correct_info = False

    if rat_columns_info["Red"]["type"] != gdal.GFT_Integer:
        correct_info = False
    if rat_columns_info["Red"]["usage"] != gdal.GFU_Red:
        correct_info = False

    if rat_columns_info["b1Mean"]["type"] != gdal.GFT_Real:
        correct_info = False
    if rat_columns_info["b1Mean"]["usage"] != gdal.GFU_Generic:
        correct_info = False

    assert correct_info


def test_populate_rat_with_stats(tmp_path):
    import rsgislib.rastergis
    import numpy

    base_clumps_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(base_clumps_img, clumps_img)

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    band_stats = list()
    band_stats.append(
        rsgislib.rastergis.BandAttStats(
            band=1,
            min_field="b1Min",
            max_field="b1Max",
            sum_field="b1Sum",
            std_dev_field="b1StdDev",
            mean_field="b1Mean",
        )
    )
    band_stats.append(
        rsgislib.rastergis.BandAttStats(
            band=2,
            min_field="b2Min",
            max_field="b2Max",
            sum_field="b2Sum",
            std_dev_field="b2StdDev",
            mean_field="b2Mean",
        )
    )
    band_stats.append(
        rsgislib.rastergis.BandAttStats(
            band=3,
            min_field="b3Min",
            max_field="b3Max",
            sum_field="b3Sum",
            std_dev_field="b3StdDev",
            mean_field="b3Mean",
        )
    )
    band_stats.append(
        rsgislib.rastergis.BandAttStats(
            band=4,
            min_field="b4Min",
            max_field="b4Max",
            sum_field="b4Sum",
            std_dev_field="b4StdDev",
            mean_field="b4Mean",
        )
    )

    rsgislib.rastergis.populate_rat_with_stats(input_img, clumps_img, band_stats)

    ref_clumps_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_attref.kea"
    )

    vars_to_test = [
        "Histogram",
        "b1Min",
        "b2Max",
        "b3Sum",
        "b4StdDev",
        "b1Mean",
        "b2Mean",
        "b3Mean",
        "b4Mean",
    ]
    vars_eq_vals = True
    for var in vars_to_test:
        print("Testing: {}".format(var))

        ref_vals = rsgislib.rastergis.get_column_data(ref_clumps_img, var)
        calcd_vals = rsgislib.rastergis.get_column_data(clumps_img, var)
        if calcd_vals.shape[0] != ref_vals.shape[0]:
            vars_eq_vals = False
            break
        if numpy.sum(calcd_vals) != numpy.sum(ref_vals):
            vars_eq_vals = False
            break

    assert vars_eq_vals


def test_pop_rat_img_stats(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_nostats.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_nostats.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.pop_rat_img_stats(
        clumps_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
    )


def test_collapse_rat(tmp_path):
    import rsgislib.rastergis

    clumps_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_cls_out.kea"
    )
    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.rastergis.collapse_rat(clumps_img, "OutClassName", output_img, "KEA")

    assert os.path.exists(output_img)


def test_calc_border_length(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.calc_border_length(clumps_img, "out_col", True)


def test_calc_rel_border(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_cls_out.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_cls_out.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.calc_rel_border(
        clumps_img, "rel_border_forest", "OutClassName", "Forest", True
    )


@pytest.mark.skipif(
    True,
    reason="Sometimes a seg fault with calc_rel_diff_neigh_stats which haven't found fix for yet.",
)
def test_calc_rel_diff_neigh_stats(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_cls_out.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_cls_out.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.find_neighbours(clumps_img, 1)

    fieldInfo = rsgislib.rastergis.FieldAttStats(
        field="b8_mean", min_field="MinNIRMeanDiff", max_field="MaxNIRMeanDiff"
    )
    rsgislib.rastergis.calc_rel_diff_neigh_stats(clumps_img, fieldInfo, False, 1)


@pytest.mark.skipif(
    True,
    reason="Sometimes a seg fault with calc_rel_diff_neigh_stats which haven't found fix for yet.",
)
def test_calc_rel_diff_neigh_stats_abs(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_cls_out.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_cls_out.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.find_neighbours(clumps_img, 1)

    fieldInfo = rsgislib.rastergis.FieldAttStats(
        field="b8_mean", min_field="MinNIRMeanDiff", max_field="MaxNIRMeanDiff"
    )
    rsgislib.rastergis.calc_rel_diff_neigh_stats(clumps_img, fieldInfo, True, 1)


def test_define_border_clumps(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.define_border_clumps(clumps_img, "out_col")


# TODO rsgislib.rastergis.define_clump_tile_positions


def test_find_boundary_pixels(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.rastergis.find_boundary_pixels(clumps_img, output_img, "KEA", 1)

    assert os.path.exists(output_img)


def test_find_neighbours(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.find_neighbours(clumps_img, 1)


# TODO rsgislib.rastergis.populate_rat_with_cat_proportions


def test_populate_rat_with_percentiles(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")

    band_percents = []
    band_percents.append(
        rsgislib.rastergis.BandAttPercentiles(percentile=25.0, field_name="B1Per25")
    )
    band_percents.append(
        rsgislib.rastergis.BandAttPercentiles(percentile=50.0, field_name="B1Per50")
    )
    band_percents.append(
        rsgislib.rastergis.BandAttPercentiles(percentile=75.0, field_name="B1Per75")
    )
    rsgislib.rastergis.populate_rat_with_percentiles(
        input_img, clumps_img, 1, band_percents
    )


def test_populate_rat_with_meanlit_stats(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_cls_out.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_cls_out.kea")
    copy2(input_ref_img, clumps_img)

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    input_ndvi_img = os.path.join(IMGCALC_DATA_DIR, "sen2_20210527_aber_ndvi.kea")

    band_stats = list()
    band_stats.append(rsgislib.rastergis.BandAttStats(band=1, mean_field="b1_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=2, mean_field="b2_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=3, mean_field="b3_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=4, mean_field="b4_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=5, mean_field="b5_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=6, mean_field="b6_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=7, mean_field="b7_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=8, mean_field="b8_meanlit"))
    band_stats.append(rsgislib.rastergis.BandAttStats(band=9, mean_field="b9_meanlit"))
    band_stats.append(
        rsgislib.rastergis.BandAttStats(band=10, mean_field="b10_meanlit")
    )

    rsgislib.rastergis.populate_rat_with_meanlit_stats(
        input_img,
        clumps_img,
        input_ndvi_img,
        1,
        "ndvi_mean",
        "meanlit_pxl_ct",
        band_stats,
    )


# TODO rsgislib.rastergis.select_clumps_on_grid


def test_clumps_spatial_location(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.clumps_spatial_location(
        clumps_img, eastings="eastings", northings="northings"
    )


def test_clumps_spatial_extent(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.clumps_spatial_extent(
        clumps_img,
        min_xx="min_xx",
        min_xy="min_xy",
        max_xx="max_xx",
        max_xy="max_xy",
        min_yx="min_yx",
        min_yy="min_yy",
        max_yx="max_yx",
        max_yy="max_yy",
        rat_band=1,
    )


def test_populate_rat_with_mode(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    in_cls_img = os.path.join(DATA_DIR, "sen2_20210527_aber_cls.kea")

    rsgislib.rastergis.populate_rat_with_mode(
        in_cls_img,
        clumps_img,
        out_cols_name="cls_val",
        use_no_data=True,
        no_data_val=0,
        out_no_data=0,
        mode_band=1,
        rat_band=1,
    )


def test_populate_rat_with_prop_valid_pxls(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    in_cls_img = os.path.join(DATA_DIR, "sen2_20210527_aber_cls.kea")

    rsgislib.rastergis.populate_rat_with_prop_valid_pxls(
        in_cls_img, clumps_img, out_col="cls_val", no_data_val=0, rat_band=1
    )


def test_export_col_to_gdal_img(tmp_path):
    import rsgislib.rastergis

    clumps_img = os.path.join(
        CLASSIFICATION_DATA_DIR, "sen2_20210527_aber_clumps_s2means.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.rastergis.export_col_to_gdal_img(
        clumps_img, output_img, "KEA", rsgislib.TYPE_32FLOAT, "b6Mean", rat_band=1
    )

    assert os.path.exists(output_img)


def test_export_cols_to_gdal_img(tmp_path):
    import rsgislib.rastergis

    clumps_img = os.path.join(
        CLASSIFICATION_DATA_DIR, "sen2_20210527_aber_clumps_s2means.kea"
    )

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.rastergis.export_cols_to_gdal_img(
        clumps_img,
        output_img,
        "KEA",
        rsgislib.TYPE_32FLOAT,
        ["b1Mean", "b2Mean", "b3Mean"],
        rat_band=1,
    )

    assert os.path.exists(output_img)


def test_export_rat_cols_to_ascii(tmp_path):
    import rsgislib.rastergis

    clumps_img = os.path.join(
        CLASSIFICATION_DATA_DIR, "sen2_20210527_aber_clumps_s2means.kea"
    )

    out_file = os.path.join(tmp_path, "out_data.txt")
    rsgislib.rastergis.export_rat_cols_to_ascii(
        clumps_img, out_file, ["b1Mean", "b2Mean", "b3Mean"], rat_band=1
    )

    assert os.path.exists(out_file)


def test_export_clumps_to_images(tmp_path):
    import rsgislib.rastergis
    import glob

    clumps_img = os.path.join(RASTERGIS_DATA_DIR, "sen2_grid_clumps.kea")
    out_img_base = os.path.join(tmp_path, "out_img_")

    rsgislib.rastergis.export_clumps_to_images(
        clumps_img, out_img_base, True, "kea", "KEA", rat_band=1
    )

    assert len(glob.glob("{}*.kea".format(out_img_base))) == 4


def test_get_column_data():
    import rsgislib.rastergis
    import numpy

    ref_clumps_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_attref.kea"
    )

    hist_col_vals = rsgislib.rastergis.get_column_data(ref_clumps_img, "Histogram")

    hist_vals_range_ok = False
    if (
        (numpy.min(hist_col_vals) >= 0)
        and (numpy.max(hist_col_vals) <= 80174)
        and (hist_col_vals.shape[0] == 11949)
    ):
        hist_col_vals = True

    assert hist_col_vals


def test_set_column_data(tmp_path):
    import rsgislib.rastergis
    import numpy

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    n_rows = rsgislib.rastergis.get_rat_length(clumps_img)
    uid_col = numpy.arange(0, n_rows, 1, dtype=numpy.uint32)
    rsgislib.rastergis.set_column_data(clumps_img, "test_col", uid_col)

    read_col_vals = rsgislib.rastergis.get_column_data(clumps_img, "test_col")
    assert numpy.array_equal(read_col_vals, uid_col)


def test_create_uid_col(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_clumps.kea")
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps.kea")
    copy2(input_ref_img, clumps_img)

    rsgislib.rastergis.create_uid_col(clumps_img)


def test_get_global_class_stats(tmp_path):
    import rsgislib.rastergis

    input_ref_img = os.path.join(
        RASTERGIS_DATA_DIR, "sen2_20210527_aber_clumps_cls_out.kea"
    )
    clumps_img = os.path.join(tmp_path, "sen2_20210527_aber_clumps_cls_out.kea")
    copy2(input_ref_img, clumps_img)

    change_feat_vals = list()
    change_feat_vals.append(rsgislib.rastergis.ChangeFeats(cls_name="Forest"))
    change_feat_vals.append(rsgislib.rastergis.ChangeFeats(cls_name="Grass"))

    rsgislib.rastergis.get_global_class_stats(
        clumps_img, "OutClassName", ["ndvi_mean"], change_feat_vals
    )


# TODO rsgislib.rastergis.str_class_majority
# TODO rsgislib.rastergis.histo_sampling
# TODO rsgislib.rastergis.class_split_fit_hist_gausian_mixture_model
# TODO rsgislib.rastergis.apply_rat_knn
# TODO rsgislib.rastergis.get_global_class_stats
# TODO rsgislib.rastergis.fit_hist_gausian_mixture_model
# TODO rsgislib.rastergis.calc_1d_jm_distance
# TODO rsgislib.rastergis.calc_2d_jm_distance
# TODO rsgislib.rastergis.calc_bhattacharyya_distance
# TODO rsgislib.rastergis.copy_gdal_rat_columns
# TODO rsgislib.rastergis.copy_rat
# TODO rsgislib.rastergis.import_vec_atts
# TODO rsgislib.rastergis.colour_rat_classes
# TODO rsgislib.rastergis.define_class_names
# TODO rsgislib.rastergis.take_random_sample
# TODO rsgislib.rastergis.set_class_names_colours
# TODO rsgislib.rastergis.calc_dist_between_clumps
# TODO rsgislib.rastergis.calc_dist_to_large_clumps
# TODO rsgislib.rastergis.calc_dist_to_classes
# TODO rsgislib.rastergis.identify_small_units



#! /usr/bin/env python
############################################################################
#  pxloutlierchng.py
#
#  Copyright 2019 RSGISLib.
#
#  RSGISLib: 'The remote sensing and GIS Software Library'
#
#  RSGISLib is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  RSGISLib is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with RSGISLib.  If not, see <http://www.gnu.org/licenses/>.
#
#
# Purpose:  Provide a set of utilities to perform per-pixel change detection
#           using outlier methods. These are sometime called map-to-image
#           methods of change detection.
#
# Author: Pete Bunting
# Email: petebunting@mac.com
# Date: 12/06/2019
# Version: 1.0
#
# History:
# Version 1.0 - Created.
#
###########################################################################

from __future__ import print_function

import numpy

from rios import applier
from rios import cuiprogress

import rsgislib
import rsgislib.rastergis
import rsgislib.imageutils
import rsgislib.imagecalc


def find_class_pyod_outliers(
    pyod_obj,
    input_img,
    in_msk_img,
    output_img,
    out_scores_img=None,
    img_mask_val=1,
    img_bands=None,
    gdalformat="KEA",
):
    """
    This function uses the pyod (https://github.com/yzhao062/pyod) library to
    find outliers within a class. It is assumed that the input images are from a
    different date than the mask (classification) and therefore the outliers will
    related to class changes.

    :param pyod_obj: an instance of a pyod.models (e.g., pyod.models.knn.KNN) pass
                     parameters to the constructor
    :param input_img: input image used for analysis
    :param in_msk_img: input image mask use to define the region of interest.
    :param output_img: output image with pixel over of 1 for within mask but
                         not outlier and 2 for in mask and outlier.
    :param out_scores_img: output image (optional, None and won't be provided;
                           Default None) providing the probability of
                           each pixel being an outlier
    :param img_mask_val: the pixel value within the mask image for the
                         class of interest. (Default 1)
    :param img_bands: the image bands to be used for the analysis.
                      If None then all used (Default: None)
    :param gdalformat: file format for the output image(s). Default KEA.

    """
    if img_bands is not None:
        if not ((type(img_bands) is list) or (type(img_bands) is tuple)):
            raise rsgislib.RSGISPyException(
                "If provided then img_bands should be a list (or None)"
            )
    else:
        n_bands = rsgislib.imageutils.get_img_band_count(input_img)
        img_bands = numpy.arange(1, n_bands + 1)
    num_vars = len(img_bands)
    img_val_no_data = rsgislib.imageutils.get_img_no_data_value(input_img)

    msk_arr_vals = rsgislib.imageutils.extract_img_pxl_vals_in_msk(
        input_img, img_bands, in_msk_img, img_mask_val, img_val_no_data
    )
    print("There were {} pixels within the mask.".format(msk_arr_vals.shape[0]))

    print("Fitting outlier detector")
    pyod_obj.fit(msk_arr_vals)
    print("Fitted outlier detector")

    # RIOS function to apply classifier
    def _applyPyOB(info, inputs, outputs, otherargs):
        # Internal function for rios applier. Used within find_class_outliers.

        out_lbls_vals = numpy.zeros_like(inputs.image_mask, dtype=numpy.uint8)
        if otherargs.out_scores:
            out_scores_vals = numpy.zeros_like(inputs.image_mask, dtype=numpy.float32)
        if numpy.any(inputs.image_mask == otherargs.msk_val):
            out_lbls_vals = out_lbls_vals.flatten()
            img_msk_vals = inputs.image_mask.flatten()
            if otherargs.out_scores:
                out_scores_vals = out_scores_vals.flatten()
            ID = numpy.arange(img_msk_vals.shape[0])

            img_shape = inputs.input_img.shape
            img_bands = inputs.input_img.reshape(
                (img_shape[0], (img_shape[1] * img_shape[2]))
            )

            band_lst = []
            for band in otherargs.img_bands:
                if (band > 0) and (band <= img_shape[0]):
                    band_lst.append(img_bands[band - 1])
                else:
                    raise rsgislib.RSGISPyException(
                        "Band ({}) specified is not within the image".format(band)
                    )
            img_bands_sel = numpy.stack(band_lst, axis=0)
            img_bands_trans = numpy.transpose(img_bands_sel)

            if otherargs.no_data_val is not None:
                ID = ID[(img_bands_trans != otherargs.no_data_val).all(axis=1)]
                img_msk_vals = img_msk_vals[
                    (img_bands_trans != otherargs.no_data_val).all(axis=1)
                ]
                img_bands_trans = img_bands_trans[
                    (img_bands_trans != otherargs.no_data_val).all(axis=1)
                ]

            ID = ID[img_msk_vals == otherargs.msk_val]
            img_bands_trans = img_bands_trans[img_msk_vals == otherargs.msk_val]

            if img_bands_trans.shape[0] > 0:
                pred_lbls = otherargs.pyod_obj.predict(img_bands_trans)
                pred_lbls = pred_lbls + 1
                out_lbls_vals[ID] = pred_lbls
            out_lbls_vals = numpy.expand_dims(
                out_lbls_vals.reshape(
                    (inputs.image_mask.shape[1], inputs.image_mask.shape[2])
                ),
                axis=0,
            )

            if otherargs.out_scores:
                if img_bands_trans.shape[0] > 0:
                    pred_probs = otherargs.pyod_obj.predict_proba(
                        img_bands_trans, method="unify"
                    )
                    out_scores_vals[ID] = pred_probs[:, 1]
                out_scores_vals = numpy.expand_dims(
                    out_scores_vals.reshape(
                        (inputs.image_mask.shape[1], inputs.image_mask.shape[2])
                    ),
                    axis=0,
                )

        outputs.out_lbls_img = out_lbls_vals
        if otherargs.out_scores:
            outputs.out_scores_img = out_scores_vals

    infiles = applier.FilenameAssociations()
    infiles.image_mask = in_msk_img
    infiles.input_img = input_img

    otherargs = applier.OtherInputs()
    otherargs.pyod_obj = pyod_obj
    otherargs.msk_val = img_mask_val
    otherargs.img_bands = img_bands
    otherargs.out_scores = False
    otherargs.no_data_val = img_val_no_data

    outfiles = applier.FilenameAssociations()
    outfiles.out_lbls_img = output_img
    if out_scores_img is not None:
        outfiles.out_scores_img = out_scores_img
        otherargs.out_scores = True

    try:
        import tqdm

        progress_bar = rsgislib.TQDMProgressBar()
    except:
        progress_bar = cuiprogress.GDALProgressBar()

    aControls = applier.ApplierControls()
    aControls.progress = progress_bar
    aControls.drivername = gdalformat
    aControls.omitPyramids = True
    aControls.calcStats = False
    print("Applying the Outlier Detector")
    applier.apply(_applyPyOB, infiles, outfiles, otherargs, controls=aControls)
    print("Completed")

    if gdalformat == "KEA":
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )
        class_info_dict = dict()
        class_info_dict[1] = {"classname": "no_chng", "red": 0, "green": 255, "blue": 0}
        class_info_dict[2] = {"classname": "chng", "red": 255, "green": 0, "blue": 0}
        rsgislib.rastergis.set_class_names_colours(
            output_img, "chng_cls", class_info_dict
        )
    else:
        rsgislib.imageutils.pop_thmt_img_stats(
            output_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )

    if out_scores_img is not None:
        rsgislib.imageutils.pop_img_stats(
            out_scores_img, use_no_data=True, no_data_val=0, calc_pyramids=True
        )


def find_class_kurt_skew_outliers(
    input_img: str,
    in_msk_img: str,
    output_img: str,
    vld_min: float,
    vld_max: float,
    init_thres: float,
    low_thres: bool,
    contamination: float = 10.0,
    only_kurtosis: bool = False,
    img_mask_val: int = 1,
    img_band: int = 1,
    img_val_no_data: float = None,
    gdalformat: str = "KEA",
    plot_thres_file: str = None,
) -> float:
    """
    This function to find outliers within a class using an optimisation of the skewness
    and kurtosis. It is assumed that the input_img is from a different date than
    the mask (classification) and therefore the outliers will related to class changes.
    The skewness and kurtosis method assume that without change image values will
    be normally distributed.

    :param input_img: the input image for the analysis. Just a single band will be used.
    :param in_msk_img: input image mask use to define the region (class) of interest.
    :param output_img:  output image with pixel over of 1 for within mask but
                         not outlier and 2 for in mask and outlier.
    :param vld_min: the minimum threshold for valid image values.
    :param vld_max: the maximum threshold for valid image values.
    :param init_thres: an initial estimate of the change threshold
    :param low_thres: a boolean as to whether the threshold is on the upper or lower
                      side of the histogram. If True (default) then outliers will be
                      identified as values below the threshold. If False then outliers
                      will be above the threshold.
    :param contamination: An estimate of the amount of contamination (i.e., outliners)
                          which is within the scene.
    :param only_kurtosis: A boolean to specify that only the kurtosis should be used
                          to estimate the threshold. Default: False (i.e., both the
                        kurtosis and skewness are used.
    :param img_mask_val: the pixel value within the in_msk_img specifying the class
                         of interest.
    :param img_band: the input_img image band to be used for the analysis.
    :param img_val_no_data: the input_img image not data value. If None then the value
                            will be read from the image header.
    :param gdalformat: the output image file format. (Default: KEA)
    :param plot_thres_file: A file path for a plot of the histogram with the
                            threshold. If None then ignored.
    :return: The threshold identified.

    """
    import rsgislib.tools.stats

    if img_val_no_data is None:
        img_val_no_data = rsgislib.imageutils.get_img_no_data_value(input_img)

    msk_arr_vals = rsgislib.imageutils.extract_img_pxl_vals_in_msk(
        input_img, [img_band], in_msk_img, img_mask_val, img_val_no_data
    )
    print("There were {} pixels within the mask.".format(msk_arr_vals.shape[0]))

    chng_thres = rsgislib.tools.stats.calc_kurt_skew_threshold(
        msk_arr_vals[..., 0],
        vld_max,
        vld_min,
        init_thres,
        low_thres,
        contamination,
        only_kurtosis,
    )

    band_defns = list()
    band_defns.append(rsgislib.imagecalc.BandDefn("msk", in_msk_img, 1))
    band_defns.append(rsgislib.imagecalc.BandDefn("val", input_img, img_band))
    if low_thres:
        exp = (
            f"(val=={img_val_no_data})?0:(msk=={img_mask_val})&&"
            f"(val<{chng_thres})?2:(msk=={img_mask_val})?1:0"
        )
    else:
        exp = (
            f"(val=={img_val_no_data})?0:(msk=={img_mask_val})&&"
            f"(val>{chng_thres})?2:(msk=={img_mask_val})?1:0"
        )
    rsgislib.imagecalc.band_math(
        output_img, exp, gdalformat, rsgislib.TYPE_8UINT, band_defns
    )

    if gdalformat == "KEA":
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )
        class_info_dict = dict()
        class_info_dict[1] = {"classname": "no_chng", "red": 0, "green": 255, "blue": 0}
        class_info_dict[2] = {"classname": "chng", "red": 255, "green": 0, "blue": 0}
        rsgislib.rastergis.set_class_names_colours(
            output_img, "chng_cls", class_info_dict
        )
    else:
        rsgislib.imageutils.pop_thmt_img_stats(
            output_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )

    if plot_thres_file is not None:
        import rsgislib.tools.plotting

        rsgislib.tools.plotting.plot_histogram_threshold(
            msk_arr_vals[..., 0], plot_thres_file, chng_thres
        )

    return chng_thres


def find_class_otsu_outliers(
    input_img: str,
    in_msk_img: str,
    output_img: str,
    low_thres: bool,
    img_mask_val: int = 1,
    img_band: int = 1,
    img_val_no_data: float = None,
    gdalformat: str = "KEA",
    plot_thres_file: str = None,
) -> float:
    """
    This function to find outliers within a class using an otsu thresholding. It is
    assumed that the input_img is from a different date than the mask
    (classification) and therefore the outliers will related to class changes.

    :param input_img: the input image for the analysis. Just a single band will be used.
    :param in_msk_img: input image mask use to define the region (class) of interest.
    :param output_img:  output image with pixel over of 1 for within mask but
                         not outlier and 2 for in mask and outlier.
    :param low_thres: a boolean as to whether the threshold is on the upper or lower
                      side of the histogram. If True (default) then outliers will be
                      identified as values below the threshold. If False then outliers
                      will be above the threshold.
    :param img_mask_val: the pixel value within the in_msk_img specifying the class
                         of interest.
    :param img_band: the input_img image band to be used for the analysis.
    :param img_val_no_data: the input_img image not data value. If None then the value
                            will be read from the image header.
    :param gdalformat: the output image file format. (Default: KEA)
    :param plot_thres_file: A file path for a plot of the histogram with the
                            threshold. If None then ignored.
    :return: The threshold identified.

    """
    import rsgislib.tools.stats

    if img_val_no_data is None:
        img_val_no_data = rsgislib.imageutils.get_img_no_data_value(input_img)

    msk_arr_vals = rsgislib.imageutils.extract_img_pxl_vals_in_msk(
        input_img, [img_band], in_msk_img, img_mask_val, img_val_no_data
    )
    print("There were {} pixels within the mask.".format(msk_arr_vals.shape[0]))

    chng_thres = rsgislib.tools.stats.calc_otsu_threshold(msk_arr_vals[..., 0])

    band_defns = list()
    band_defns.append(rsgislib.imagecalc.BandDefn("msk", in_msk_img, 1))
    band_defns.append(rsgislib.imagecalc.BandDefn("val", input_img, img_band))
    if low_thres:
        exp = (
            f"(val=={img_val_no_data})?0:(msk=={img_mask_val})&&"
            f"(val<{chng_thres})?2:(msk=={img_mask_val})?1:0"
        )
    else:
        exp = (
            f"(val=={img_val_no_data})?0:(msk=={img_mask_val})&&"
            f"(val>{chng_thres})?2:(msk=={img_mask_val})?1:0"
        )
    rsgislib.imagecalc.band_math(
        output_img, exp, gdalformat, rsgislib.TYPE_8UINT, band_defns
    )

    if gdalformat == "KEA":
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )
        class_info_dict = dict()
        class_info_dict[1] = {"classname": "no_chng", "red": 0, "green": 255, "blue": 0}
        class_info_dict[2] = {"classname": "chng", "red": 255, "green": 0, "blue": 0}
        rsgislib.rastergis.set_class_names_colours(
            output_img, "chng_cls", class_info_dict
        )
    else:
        rsgislib.imageutils.pop_thmt_img_stats(
            output_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )

    if plot_thres_file is not None:
        import rsgislib.tools.plotting

        rsgislib.tools.plotting.plot_histogram_threshold(
            msk_arr_vals[..., 0], plot_thres_file, chng_thres
        )

    return chng_thres


def find_class_li_outliers(
    input_img: str,
    in_msk_img: str,
    output_img: str,
    low_thres: bool,
    tolerance: float = None,
    init_thres: float = None,
    img_mask_val: int = 1,
    img_band: int = 1,
    img_val_no_data: float = None,
    gdalformat: str = "KEA",
    plot_thres_file: str = None,
) -> float:
    """
    This function to find outliers within a class using Li's iterative Minimum Cross
    Entropy method. It is assumed that the input_img is from a different date than
    the mask (classification) and therefore the outliers will related to class changes.

    :param input_img: the input image for the analysis. Just a single band will be used.
    :param in_msk_img: input image mask use to define the region (class) of interest.
    :param output_img:  output image with pixel over of 1 for within mask but
                         not outlier and 2 for in mask and outlier.
    :param low_thres: a boolean as to whether the threshold is on the upper or lower
                      side of the histogram. If True (default) then outliers will be
                      identified as values below the threshold. If False then outliers
                      will be above the threshold.
    :param tolerance: float (optional) - Finish the computation when the
                      change in the threshold in an iteration is less than
                      this value. By default, this is half the smallest
                      difference between data values.
    :param init_thres: an initial estimate of the change threshold
    :param img_mask_val: the pixel value within the in_msk_img specifying the class
                         of interest.
    :param img_band: the input_img image band to be used for the analysis.
    :param img_val_no_data: the input_img image not data value. If None then the value
                            will be read from the image header.
    :param gdalformat: the output image file format. (Default: KEA)
    :param plot_thres_file: A file path for a plot of the histogram with the
                            threshold. If None then ignored.
    :return: The threshold identified.

    """
    import rsgislib.tools.stats

    if img_val_no_data is None:
        img_val_no_data = rsgislib.imageutils.get_img_no_data_value(input_img)

    msk_arr_vals = rsgislib.imageutils.extract_img_pxl_vals_in_msk(
        input_img, [img_band], in_msk_img, img_mask_val, img_val_no_data
    )
    print("There were {} pixels within the mask.".format(msk_arr_vals.shape[0]))

    chng_thres = rsgislib.tools.stats.calc_li_threshold(
        msk_arr_vals[..., 0],
        tolerance,
        init_thres,
    )

    band_defns = list()
    band_defns.append(rsgislib.imagecalc.BandDefn("msk", in_msk_img, 1))
    band_defns.append(rsgislib.imagecalc.BandDefn("val", input_img, img_band))
    if low_thres:
        exp = (
            f"(val=={img_val_no_data})?0:(msk=={img_mask_val})&&"
            f"(val<{chng_thres})?2:(msk=={img_mask_val})?1:0"
        )
    else:
        exp = (
            f"(val=={img_val_no_data})?0:(msk=={img_mask_val})&&"
            f"(val>{chng_thres})?2:(msk=={img_mask_val})?1:0"
        )
    rsgislib.imagecalc.band_math(
        output_img, exp, gdalformat, rsgislib.TYPE_8UINT, band_defns
    )

    if gdalformat == "KEA":
        rsgislib.rastergis.pop_rat_img_stats(
            clumps_img=output_img,
            add_clr_tab=True,
            calc_pyramids=True,
            ignore_zero=True,
        )
        class_info_dict = dict()
        class_info_dict[1] = {"classname": "no_chng", "red": 0, "green": 255, "blue": 0}
        class_info_dict[2] = {"classname": "chng", "red": 255, "green": 0, "blue": 0}
        rsgislib.rastergis.set_class_names_colours(
            output_img, "chng_cls", class_info_dict
        )
    else:
        rsgislib.imageutils.pop_thmt_img_stats(
            output_img, add_clr_tab=True, calc_pyramids=True, ignore_zero=True
        )

    if plot_thres_file is not None:
        import rsgislib.tools.plotting

        rsgislib.tools.plotting.plot_histogram_threshold(
            msk_arr_vals[..., 0], plot_thres_file, chng_thres
        )

    return chng_thres



import os

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
IMGREG_DATA_DIR = os.path.join(DATA_DIR, "imageregistration")


def test_find_image_offset():
    import rsgislib.imageregistration

    in_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    in_float_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset.kea"
    )
    x_off, y_off = rsgislib.imageregistration.find_image_offset(
        in_ref_img,
        in_float_img,
        [1, 2, 3],
        [1, 2, 3],
        rsgislib.imageregistration.METRIC_CORELATION,
        4,
        4,
        4,
    )
    print("x_off: {}".format(x_off))
    print("y_off: {}".format(y_off))
    assert abs((x_off - 3) < 0.5) and abs((y_off - 3) < 0.5)


def test_apply_offset_to_image(tmp_path):
    import rsgislib.imageregistration

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageregistration.apply_offset_to_image(
        input_img, output_img, "KEA", rsgislib.TYPE_16UINT, 50, 30
    )
    assert os.path.exists(output_img)


def test_basic_registration(tmp_path):
    import rsgislib.imageregistration

    in_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    in_float_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset.kea"
    )
    out_gcp_file = os.path.join(tmp_path, "out_gcps.txt")
    rsgislib.imageregistration.basic_registration(
        in_ref_img,
        in_float_img,
        out_gcp_file,
        50,
        0.8,
        50,
        4,
        2,
        2,
        4,
        rsgislib.imageregistration.METRIC_CORELATION,
        rsgislib.imageregistration.TYPE_RSGIS_IMG2MAP,
    )

    assert os.path.exists(out_gcp_file)


def test_single_layer_registration(tmp_path):
    import rsgislib.imageregistration

    in_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    in_float_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset.kea"
    )
    out_gcp_file = os.path.join(tmp_path, "out_gcps.txt")
    rsgislib.imageregistration.single_layer_registration(
        in_ref_img,
        in_float_img,
        out_gcp_file,
        50,
        0.8,
        50,
        4,
        2,
        2,
        5,
        6,
        4,
        0.5,
        2,
        rsgislib.imageregistration.METRIC_CORELATION,
        rsgislib.imageregistration.TYPE_RSGIS_IMG2MAP,
    )
    assert os.path.exists(out_gcp_file)


def test_gcp_to_gdal(tmp_path):
    import rsgislib.imageregistration
    import rsgislib.imageutils

    input_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset.kea"
    )
    in_gcp_file = os.path.join(IMGREG_DATA_DIR, "reg_gcps.txt")
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageregistration.gcp_to_gdal(
        input_img, in_gcp_file, output_img, "KEA", rsgislib.TYPE_16UINT
    )
    assert os.path.exists(output_img) and rsgislib.imageutils.has_gcps(output_img)


def test_warp_with_gcps_with_gdal(tmp_path):
    import rsgislib.imageregistration

    in_ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_b123.kea")
    in_process_img = os.path.join(
        IMGREG_DATA_DIR, "sen2_20210527_aber_subset_b123_offset_gcps.kea"
    )
    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageregistration.warp_with_gcps_with_gdal(
        in_ref_img,
        in_process_img,
        output_img,
        "KEA",
        rsgislib.INTERP_NEAREST_NEIGHBOUR,
        use_tps=False,
        use_poly=True,
        poly_order=2,
        use_multi_thread=False,
    )



#!/usr/bin/env python
"""
The vector attributes module performs attribute table operations on vectors.
"""

from typing import Union, List, Dict
import math

from osgeo import gdal
import osgeo.ogr
from osgeo import ogr

import numpy

import rsgislib

gdal.UseExceptions()


def write_vec_column(
    out_vec_file: str,
    out_vec_lyr: str,
    att_column: str,
    att_col_datatype: int,
    att_col_data: List,
):
    """
    A function which will write a column to a vector file

    :param out_vec_file: The file / path to the vector data 'file'.
    :param out_vec_lyr: The layer to which the data is to be added.
    :param att_column: Name of the output column
    :param att_col_datatype: ogr data type for the output column
                            (e.g., ogr.OFTString, ogr.OFTInteger, ogr.OFTReal)
    :param att_col_data: A list of the same length as the number of features
                        in vector file.

    """
    ds = gdal.OpenEx(out_vec_file, gdal.OF_UPDATE)
    if ds is None:
        raise rsgislib.RSGISPyException("Could not open '{}'".format(out_vec_file))

    lyr = ds.GetLayerByName(out_vec_lyr)
    if lyr is None:
        raise rsgislib.RSGISPyException("Could not find layer '{}'".format(out_vec_lyr))

    numFeats = lyr.GetFeatureCount()
    if not len(att_col_data) == numFeats:
        print("Number of Features: {}".format(numFeats))
        print("Length of Data: {}".format(len(att_col_data)))
        raise rsgislib.RSGISPyException(
            "The number of features and size of the input data is not equal."
        )

    colExists = False
    lyrDefn = lyr.GetLayerDefn()
    for i in range(lyrDefn.GetFieldCount()):
        if lyrDefn.GetFieldDefn(i).GetName().lower() == att_column.lower():
            colExists = True
            break

    if not colExists:
        field_defn = ogr.FieldDefn(att_column, att_col_datatype)
        if lyr.CreateField(field_defn) != 0:
            raise rsgislib.RSGISPyException(
                "Creating '{}' field failed; be careful with case, "
                "some drivers are case insensitive but column might "
                "not be found.".format(att_column)
            )

    lyr.ResetReading()
    # WORK AROUND AS SQLITE GETS STUCK IN LOOP ON FIRST FEATURE WHEN USE SETFEATURE.
    fids = []
    for feat in lyr:
        fids.append(feat.GetFID())

    openTransaction = False
    lyr.ResetReading()
    i = 0
    try:
        # WORK AROUND AS SQLITE GETS STUCK IN LOOP ON FIRST FEATURE WHEN USE SETFEATURE.
        for fid in fids:
            if not openTransaction:
                lyr.StartTransaction()
                openTransaction = True
            feat = lyr.GetFeature(fid)
            if feat is not None:
                feat.SetField(att_column, att_col_data[i])
                lyr.SetFeature(feat)
            if ((i % 20000) == 0) and openTransaction:
                lyr.CommitTransaction()
                openTransaction = False
            i = i + 1
        if openTransaction:
            lyr.CommitTransaction()
            openTransaction = False
        lyr.SyncToDisk()
        ds = None
    except Exception as e:
        if i < numFeats:
            print(
                "Data type of the value being "
                "written is '{}'".format(type(att_col_data[i]))
            )
        raise e


def write_vec_column_to_layer(
    out_vec_lyr_obj: osgeo.ogr.Layer,
    att_column: str,
    att_col_datatype: int,
    att_col_data: List,
):
    """
    A function which will write a column to a vector layer.

    :param out_vec_lyr_obj: GDAL/OGR vector layer object
    :param att_column: Name of the output column
    :param att_col_datatype: ogr data type for the output column
                            (e.g., ogr.OFTString, ogr.OFTInteger, ogr.OFTReal)
    :param att_col_data: A list of the same length as the number of features
                         in vector file.

    """
    if out_vec_lyr_obj is None:
        raise rsgislib.RSGISPyException("The layer passed in is None...")

    numFeats = out_vec_lyr_obj.GetFeatureCount()
    if not len(att_col_data) == numFeats:
        print("Number of Features: {}".format(numFeats))
        print("Length of Data: {}".format(len(att_col_data)))
        raise rsgislib.RSGISPyException(
            "The number of features and size of the input data is not equal."
        )

    colExists = False
    lyrDefn = out_vec_lyr_obj.GetLayerDefn()
    for i in range(lyrDefn.GetFieldCount()):
        if lyrDefn.GetFieldDefn(i).GetName().lower() == att_column.lower():
            colExists = True
            break

    if not colExists:
        field_defn = ogr.FieldDefn(att_column, att_col_datatype)
        if out_vec_lyr_obj.CreateField(field_defn) != 0:
            raise rsgislib.RSGISPyException(
                "Creating '{}' field failed; be careful with case, some "
                "drivers are case insensitive but column might not "
                "be found.".format(att_column)
            )

    out_vec_lyr_obj.ResetReading()
    # WORK AROUND AS SQLITE GETS STUCK IN LOOP ON FIRST FEATURE WHEN USE SETFEATURE.
    fids = []
    for feat in out_vec_lyr_obj:
        fids.append(feat.GetFID())

    openTransaction = False
    out_vec_lyr_obj.ResetReading()
    i = 0
    # WORK AROUND AS SQLITE GETS STUCK IN LOOP ON FIRST FEATURE WHEN USE SETFEATURE.
    for fid in fids:
        if not openTransaction:
            out_vec_lyr_obj.StartTransaction()
            openTransaction = True
        feat = out_vec_lyr_obj.GetFeature(fid)
        if feat is not None:
            feat.SetField(att_column, att_col_data[i])
            out_vec_lyr_obj.SetFeature(feat)
        if ((i % 20000) == 0) and openTransaction:
            out_vec_lyr_obj.CommitTransaction()
            openTransaction = False
        i = i + 1
    if openTransaction:
        out_vec_lyr_obj.CommitTransaction()
        openTransaction = False


def read_vec_column(vec_file: str, vec_lyr: str, att_column: str) -> List:
    """
    A function which will reads a column from a vector file

    :param vec_file: The file / path to the vector data 'file'.
    :param vec_lyr: The layer to which the data is to be read from.
    :param att_column: Name of the input column
    :returns: a list with the column values.

    """
    ds = gdal.OpenEx(vec_file, gdal.OF_VECTOR)
    if ds is None:
        raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

    lyr = ds.GetLayerByName(vec_lyr)
    if lyr is None:
        raise rsgislib.RSGISPyException("Could not find layer '{}'".format(vec_lyr))

    colExists = False
    lyrDefn = lyr.GetLayerDefn()
    for i in range(lyrDefn.GetFieldCount()):
        if lyrDefn.GetFieldDefn(i).GetName() == att_column:
            colExists = True
            break

    if not colExists:
        ds = None
        raise rsgislib.RSGISPyException(
            "The specified column does not exist in the input layer; "
            "check case as some drivers are case sensitive."
        )

    outVal = list()
    lyr.ResetReading()
    for feat in lyr:
        outVal.append(feat.GetField(att_column))
    ds = None

    return outVal


def read_vec_columns(vec_file: str, vec_lyr: str, att_columns: List[str]) -> List[Dict]:
    """
    A function which will reads a number of column from a vector file

    :param vec_file: The file / path to the vector data 'file'.
    :param vec_lyr: The layer to which the data is to be read from.
    :param att_columns: List of input attribute column names to be read in.
    :return: list of dicts with the column names as keys

    """
    ds = gdal.OpenEx(vec_file, gdal.OF_VECTOR)
    if ds is None:
        raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

    lyr = ds.GetLayerByName(vec_lyr)
    if lyr is None:
        raise rsgislib.RSGISPyException("Could not find layer '{}'".format(vec_lyr))

    lyrDefn = lyr.GetLayerDefn()

    feat_idxs = dict()
    feat_types = dict()
    found_atts = dict()
    for attName in att_columns:
        found_atts[attName] = False

    for i in range(lyrDefn.GetFieldCount()):
        if lyrDefn.GetFieldDefn(i).GetName() in att_columns:
            attName = lyrDefn.GetFieldDefn(i).GetName()
            feat_idxs[attName] = i
            feat_types[attName] = lyrDefn.GetFieldDefn(i).GetType()
            found_atts[attName] = True

    for attName in att_columns:
        if not found_atts[attName]:
            ds = None
            raise rsgislib.RSGISPyException(
                "Could not find the attribute ({}) specified "
                "within the vector layer.".format(attName)
            )

    outvals = []
    lyr.ResetReading()
    for feat in lyr:
        outdict = dict()
        for attName in att_columns:
            if feat_types[attName] == ogr.OFTString:
                outdict[attName] = feat.GetFieldAsString(feat_idxs[attName])
            elif feat_types[attName] == ogr.OFTReal:
                outdict[attName] = feat.GetFieldAsDouble(feat_idxs[attName])
            elif feat_types[attName] == ogr.OFTInteger:
                outdict[attName] = feat.GetFieldAsInteger(feat_idxs[attName])
            else:
                outdict[attName] = feat.GetField(feat_idxs[attName])
        outvals.append(outdict)
    ds = None

    return outvals


def pop_bbox_cols(
    vec_file: str,
    vec_lyr: str,
    x_min_col: str = "xmin",
    x_max_col: str = "xmax",
    y_min_col: str = "ymin",
    y_max_col: str = "ymax",
):
    """
    A function which adds a polygons boundary bbox as attributes to each feature.

    :param vec_file: vector file.
    :param vec_lyr: layer within the vector file.
    :param x_min_col: output column name.
    :param x_max_col: output column name.
    :param y_min_col: output column name.
    :param y_max_col: output column name.

    """
    dsVecFile = gdal.OpenEx(vec_file, gdal.OF_UPDATE)
    if dsVecFile is None:
        raise rsgislib.RSGISPyException("Could not open '{}'".format(vec_file))

    vec_lyr_obj = dsVecFile.GetLayerByName(vec_lyr)
    if vec_lyr_obj is None:
        raise rsgislib.RSGISPyException("Could not find layer '{}'".format(vec_lyr))

    xminCol_exists = False
    xmaxCol_exists = False
    yminCol_exists = False
    ymaxCol_exists = False

    lyrDefn = vec_lyr_obj.GetLayerDefn()
    for i in range(lyrDefn.GetFieldCount()):
        if lyrDefn.GetFieldDefn(i).GetName() == x_min_col:
            xminCol_exists = True
        if lyrDefn.GetFieldDefn(i).GetName() == x_max_col:
            xmaxCol_exists = True
        if lyrDefn.GetFieldDefn(i).GetName() == y_min_col:
            yminCol_exists = True
        if lyrDefn.GetFieldDefn(i).GetName() == y_max_col:
            ymaxCol_exists = True

    if not xminCol_exists:
        xmin_field_defn = ogr.FieldDefn(x_min_col, ogr.OFTReal)
        if vec_lyr_obj.CreateField(xmin_field_defn) != 0:
            raise rsgislib.RSGISPyException(
                "Creating '{}' field failed.".format(x_min_col)
            )

    if not xmaxCol_exists:
        xmax_field_defn = ogr.FieldDefn(x_max_col, ogr.OFTReal)
        if vec_lyr_obj.CreateField(xmax_field_defn) != 0:
            raise rsgislib.RSGISPyException(
                "Creating '{}' field failed.".format(x_max_col)
            )

    if not yminCol_exists:
        ymin_field_defn = ogr.FieldDefn(y_min_col, ogr.OFTReal)
        if vec_lyr_obj.CreateField(ymin_field_defn) != 0:
            raise rsgislib.RSGISPyException(
                "Creating '{}' field failed.".format(y_min_col)
            )

    if not ymaxCol_exists:
        ymax_field_defn = ogr.FieldDefn(y_max_col, ogr.OFTReal)
        if vec_lyr_obj.CreateField(ymax_field_defn) != 0:
            raise rsgislib.RSGISPyException(
                "Creating '{}' field failed.".format(y_max_col)
            )

    # WORK AROUND AS SQLITE GETS STUCK IN LOOP ON FIRST FEATURE WHEN USE SETFEATURE.
    fids = []
    for feat in vec_lyr_obj:
        fids.append(feat.GetFID())

    openTransaction = False
    nFeats = vec_lyr_obj.GetFeatureCount(True)
    step = math.floor(nFeats / 10)
    feedback = 10
    feedback_next = step
    counter = 0
    print("Started .0.", end="", flush=True)
    vec_lyr_obj.ResetReading()
    for fid in fids:
        # WORK AROUND AS SQLITE GETS STUCK IN LOOP ON FIRST FEATURE WHEN USE SETFEATURE.
        feat = vec_lyr_obj.GetFeature(fid)
        if (nFeats > 10) and (counter == feedback_next):
            print(".{}.".format(feedback), end="", flush=True)
            feedback_next = feedback_next + step
            feedback = feedback + 10

        if not openTransaction:
            vec_lyr_obj.StartTransaction()
            openTransaction = True

        geom = feat.GetGeometryRef()
        if geom is not None:
            env = geom.GetEnvelope()
            feat.SetField(x_min_col, env[0])
            feat.SetField(x_max_col, env[1])
            feat.SetField(y_min_col, env[2])
            feat.SetField(y_max_col, env[3])
        else:
            feat.SetField(x_min_col, 0.0)
            feat.SetField(x_max_col, 0.0)
            feat.SetField(y_min_col, 0.0)
            feat.SetField(y_max_col, 0.0)
        rtn_val = vec_lyr_obj.SetFeature(feat)
        if rtn_val != ogr.OGRERR_NONE:
            raise rsgislib.RSGISPyException(
                "An error has occurred setting a feature on a layer."
            )
        if ((counter % 20000) == 0) and openTransaction:
            vec_lyr_obj.CommitTransaction()
            openTransaction = False
        counter = counter + 1
    if openTransaction:
        vec_lyr_obj.CommitTransaction()
        openTransaction = False
    vec_lyr_obj.SyncToDisk()
    dsVecFile = None
    print(" Completed")


def add_geom_bbox_cols(
    vec_file: str,
    vec_lyr: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str = "GPKG",
    min_x_col: str = "MinX",
    max_x_col: str = "MaxX",
    min_y_col: str = "MinY",
    max_y_col: str = "MaxY",
):
    """
    A function which adds columns to the vector layer with the bbox of each geometry.

    :param vec_file: input vector file
    :param vec_lyr: input vector layer name
    :param out_vec_file: output vector file
    :param out_vec_lyr: output vector layer name
    :param out_format: The output format of the output file. (Default: GPKG)
    :param min_x_col: Name of the MinX column (Default: MinX)
    :param max_x_col: Name of the MaxX column (Default: MaxX)
    :param min_y_col: Name of the MinY column (Default: MinY)
    :param max_y_col: Name of the MaxY column (Default: MaxY)

    """
    import geopandas

    # Read input vector file.
    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)

    # Get Geometry bounds
    geom_bounds = base_gpdf["geometry"].bounds

    # Add columns to the geodataframe
    base_gpdf[min_x_col] = geom_bounds["minx"]
    base_gpdf[max_x_col] = geom_bounds["maxx"]
    base_gpdf[min_y_col] = geom_bounds["miny"]
    base_gpdf[max_y_col] = geom_bounds["maxy"]

    # Output the file.
    if out_format == "GPKG":
        base_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        base_gpdf.to_file(out_vec_file, driver=out_format)


def create_name_col(
    vec_file: str,
    vec_lyr: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str = "GPKG",
    out_col: str = "names",
    x_col: str = "MinX",
    y_col: str = "MaxY",
    prefix: str = "",
    postfix: str = "",
    coords_lat_lon: bool = True,
    int_coords: bool = True,
    zero_x_pad: int = 0,
    zero_y_pad: int = 0,
    round_n_digts: int = 0,
    non_neg: bool = False,
):
    """
    A function which creates a column in the vector layer which can define a name
    using coordinates associated with the feature. Often this is useful if a tiling
    has been created and from this a set of images are to generated for example.

    :param vec_file: input vector file
    :param vec_lyr: input vector layer name
    :param out_vec_file: output vector file
    :param out_vec_lyr: output vector layer name
    :param out_format: The output format of the output file. (Default: GPKG)
    :param out_col: The name of the output column
    :param x_col: The column with the x coordinate
    :param y_col: The column with the y coordinate
    :param prefix: A prefix to the name
    :param postfix: A postfix to the name
    :param coords_lat_lon: A boolean specifying if the coordinates are lat / long
    :param int_coords: A boolean specifying whether to integerise the coordinates.
    :param zero_x_pad: If larger than zero then the X coordinate will be zero padded.
    :param zero_y_pad: If larger than zero then the Y coordinate will be zero padded.
    :param round_n_digts: If larger than zero then the coordinates will be rounded
                          to n significant digits
    :param non_neg: boolean specifying whether an negative coordinates should be
                    made positive. (Default: False)

    """
    import geopandas
    import tqdm
    import rsgislib.tools.utils
    import rsgislib.vectorutils

    out_format = rsgislib.vectorutils.check_format_name(out_format)

    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)

    names = list()
    for i in tqdm.tqdm(range(base_gpdf.shape[0])):
        x_col_val = base_gpdf.loc[i][x_col]
        y_col_val = base_gpdf.loc[i][y_col]

        x_col_val_neg = False
        y_col_val_neg = False
        if non_neg:
            if x_col_val < 0:
                x_col_val_neg = True
                x_col_val = x_col_val * (-1)
            if y_col_val < 0:
                y_col_val_neg = True
                y_col_val = y_col_val * (-1)

        if zero_x_pad > 0:
            x_col_val_str = rsgislib.tools.utils.zero_pad_num_str(
                x_col_val,
                str_len=zero_x_pad,
                round_num=False,
                round_n_digts=round_n_digts,
                integerise=int_coords,
            )
        else:
            x_col_val = int(x_col_val)
            x_col_val_str = "{}".format(x_col_val)

        if zero_y_pad > 0:
            y_col_val_str = rsgislib.tools.utils.zero_pad_num_str(
                y_col_val,
                str_len=zero_y_pad,
                round_num=False,
                round_n_digts=round_n_digts,
                integerise=int_coords,
            )
        else:
            y_col_val = int(y_col_val)
            y_col_val_str = "{}".format(y_col_val)

        if coords_lat_lon:
            hemi = "N"
            if y_col_val_neg:
                hemi = "S"
            east_west = "E"
            if x_col_val_neg:
                east_west = "W"
            name = "{}{}{}{}{}{}".format(
                prefix, hemi, y_col_val_str, east_west, x_col_val_str, postfix
            )
        else:
            name = "{}E{}N{}{}".format(prefix, x_col_val_str, y_col_val_str, postfix)

        names.append(name)

    base_gpdf[out_col] = numpy.array(names)

    if out_format == "GPKG":
        base_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        base_gpdf.to_file(out_vec_file, driver=out_format)


def add_unq_numeric_col(
    vec_file: str,
    vec_lyr: str,
    unq_col: str,
    out_col: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str = "GPKG",
):
    """
    A function which adds a numeric column based off an existing column in
    the vector file.

    :param vec_file: Input vector file.
    :param vec_lyr: Input vector layer within the input file.
    :param unq_col: The column within which the unique values will be identified.
    :param out_col: The output numeric column
    :param out_vec_file: Output vector file
    :param out_vec_lyr: output vector layer name.
    :param out_format: output file format (default GPKG).

    """
    import geopandas
    import rsgislib.vectorutils

    out_format = rsgislib.vectorutils.check_format_name(out_format)

    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)
    unq_vals = base_gpdf[unq_col].unique()

    base_gpdf[out_col] = numpy.zeros((base_gpdf.shape[0]), dtype=int)
    num_unq_val = 1
    for unq_val in unq_vals:
        sel_rows = base_gpdf[unq_col] == unq_val
        base_gpdf.loc[sel_rows, out_col] = num_unq_val
        num_unq_val += 1

    if out_format == "GPKG":
        base_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        base_gpdf.to_file(out_vec_file, driver=out_format)


def add_numeric_col_lut(
    vec_file: str,
    vec_lyr: str,
    ref_col: str,
    val_lut: dict,
    out_col: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str = "GPKG",
):
    """
    A function which adds a numeric column based off an existing column in the
    vector file, using an dict LUT to define the values.

    :param vec_file: Input vector file.
    :param vec_lyr: Input vector layer within the input file.
    :param ref_col: The column within which the unique values will be identified.
    :param val_lut: A dict LUT (key should be value in ref_col and value be the
                    value outputted to out_col).
    :param out_col: The output numeric column
    :param out_vec_file: Output vector file
    :param out_vec_lyr: output vector layer name.
    :param out_format: output file format (default GPKG).

    """
    import geopandas
    import rsgislib.vectorutils

    out_format = rsgislib.vectorutils.check_format_name(out_format)

    # Open vector file
    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)
    # Add output column
    base_gpdf[out_col] = numpy.zeros((base_gpdf.shape[0]), dtype=int)
    # Loop values in LUT
    for lut_key in val_lut:
        sel_rows = base_gpdf[ref_col] == lut_key
        base_gpdf.loc[sel_rows, out_col] = val_lut[lut_key]

    if out_format == "GPKG":
        base_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        base_gpdf.to_file(out_vec_file, driver=out_format)


def add_numeric_col(
    vec_file: str,
    vec_lyr: str,
    out_col: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_val: float = 1,
    out_format: str = "GPKG",
    out_col_int: bool = False,
):
    """
    A function which adds a numeric column with the same value for all the features.

    :param vec_file: Input vector file.
    :param vec_lyr: Input vector layer within the input file.
    :param out_col: The output numeric column
    :param out_vec_file: Output vector file
    :param out_vec_lyr: output vector layer name.
    :param out_val: output numeric value
    :param out_format: output file format (default GPKG).
    :param out_col_int: Specify whether the output column should be an int datatype.
                        If True (default: False) then the output column will be of
                        type int. If False then it will be type float.

    """
    import geopandas
    import rsgislib.vectorutils

    out_format = rsgislib.vectorutils.check_format_name(out_format)

    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)
    if out_col_int:
        base_gpdf[out_col] = numpy.full((base_gpdf.shape[0]), out_val, dtype=int)
    else:
        base_gpdf[out_col] = numpy.full((base_gpdf.shape[0]), out_val, dtype=float)

    if out_format == "GPKG":
        base_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        base_gpdf.to_file(out_vec_file, driver=out_format)


def add_string_col(
    vec_file: str,
    vec_lyr: str,
    out_col: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_val: str = "str_val",
    out_format: str = "GPKG",
):
    """
    A function which adds a string column with the same value for all the features.

    :param vec_file: Input vector file.
    :param vec_lyr: Input vector layer within the input file.
    :param out_col: The output numeric column
    :param out_vec_file: Output vector file
    :param out_vec_lyr: output vector layer name.
    :param out_val: output numeric value
    :param out_format: output file format (default GPKG).

    """
    import geopandas
    import rsgislib.vectorutils

    out_format = rsgislib.vectorutils.check_format_name(out_format)

    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)

    str_col = numpy.empty((base_gpdf.shape[0]), dtype=object)
    str_col[...] = out_val

    base_gpdf[out_col] = str_col

    if out_format == "GPKG":
        base_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        base_gpdf.to_file(out_vec_file, driver=out_format)


def get_unq_col_values(vec_file: str, vec_lyr: str, col_name: str) -> numpy.array:
    """
    A function which splits a vector layer by an attribute value into either
    different layers or different output files.

    :param vec_file: Input vector file
    :param vec_lyr: Input vector layer
    :param col_name: The column name for which a list of unique values will be returned.
    :returns: a numpy array as a list of the unique within the column.

    """
    import geopandas

    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)
    unq_vals = base_gpdf[col_name].unique()
    base_gpdf = None
    return unq_vals


def add_fid_col(
    vec_file: str,
    vec_lyr: str,
    out_vec_file: str,
    out_vec_lyr: str,
    out_format: str = "GPKG",
    out_col: str = "fid",
):
    """
    A function which adds a numeric feature ID (FID) column with unique values per
    feature within the file.

    :param vec_file: Input vector file.
    :param vec_lyr: Input vector layer within the input file.
    :param out_vec_file: Output vector file
    :param out_vec_lyr: output vector layer name.
    :param out_format: output file format (default GPKG).
    :param out_col: The output FID column name (Default: fid)

    """
    import geopandas
    import rsgislib.vectorutils

    out_format = rsgislib.vectorutils.check_format_name(out_format)

    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)
    base_gpdf[out_col] = numpy.arange(1, (base_gpdf.shape[0]) + 1, 1, dtype=int)

    if out_format == "GPKG":
        base_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        base_gpdf.to_file(out_vec_file, driver=out_format)


def get_vec_cols_as_array(
    vec_file: str,
    vec_lyr: str,
    cols: List[str],
    lower_limit: float = None,
    upper_limit: float = None,
) -> numpy.array:
    """
    A function returns an n x m numpy array with the values for the columns specified.

    :param vec_file: Input vector file.
    :param vec_lyr: Input vector layer within the input file.
    :param cols: list of columns to be read and returned.
    :param no_data_val: no data value used within the column values. Rows with
                        a no data value will be dropped. If None then ignored
                        (Default: None)
    :param lower_limit: Optional lower limit to define valid values. Note the same
                        value is used for all the columns listed. If a value is found
                        to be outside of the threshold the whole row is removed.
    :param upper_limit: Optional upper limit to define valid values. Note the same
                        value is used for all the columns listed. If a value is found
                        to be outside of the threshold the whole row is removed.
    :returns: a numpy array with the column values.

    """
    import geopandas
    import rsgislib.tools.stats

    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)
    sub_base_gpdf = base_gpdf.loc[:, cols]
    out_arr = sub_base_gpdf.values
    out_arr = out_arr.astype(float)

    out_arr = rsgislib.tools.stats.mask_data_to_valid(out_arr, lower_limit, upper_limit)

    return out_arr


def sort_vec_lyr(
    vec_file: str,
    vec_lyr: str,
    out_vec_file: str,
    out_vec_lyr: str,
    sort_by: Union[str, List[str]],
    ascending: Union[bool, List[bool]],
    out_format: str = "GPKG",
):
    """
    A function which sorts a vector layer based on the attributes of the layer.
    You can sort by either a single attribute or within multiple attributes
    if a list is provided. This function is implemented using geopandas.

    :param vec_file: the input vector file.
    :param vec_lyr: the input vector layer name.
    :param out_vec_file: the output vector file.
    :param out_vec_lyr: the output vector layer name.
    :param sort_by: either a string with the name of a single attribute or a list
                    of strings if multiple attributes are used for the sort.
    :param ascending: either a bool (True: ascending; False: descending) or list
                      of bools if a list of attributes was given.
    :param out_format: The output vector file format (Default: GPKG)

    """
    import geopandas

    if type(sort_by) is list:
        if type(ascending) is not list:
            raise rsgislib.RSGISPyException(
                "If sort_by is a list then ascending must be too."
            )

        if len(sort_by) != len(ascending):
            raise rsgislib.RSGISPyException(
                "If lists, the length of sort_by and ascending must be the same."
            )

    # Read input vector file.
    base_gpdf = geopandas.read_file(vec_file, layer=vec_lyr)

    # sort layer.
    sorted_gpdf = base_gpdf.sort_values(by=sort_by, ascending=ascending)

    if out_format == "GPKG":
        sorted_gpdf.to_file(out_vec_file, layer=out_vec_lyr, driver=out_format)
    else:
        sorted_gpdf.to_file(out_vec_file, driver=out_format)



#!/usr/bin/env python
"""
The tools.utils module contains some useful tools
"""

import datetime
import os
import string
import json

import numpy
import rsgislib


class RSGISNumpyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.integer):
            return int(obj)
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.ndarray):
            return obj.tolist()
        else:
            return super(RSGISNumpyArrayEncoder, self).default(obj)


def uid_generator(size: int = 6):
    """
    A function which will generate a 'random' string of the specified length
    based on the UUID

    :param size: the length of the returned string.
    :return: string of length size.

    """
    import uuid

    random_str = str(uuid.uuid4())
    random_str = random_str.replace("-", "")
    return random_str[0:size]


def is_number(str_val: str):
    """
    A function which tests whether the input string contains a number of not.

    :return: boolean

    """
    try:
        float(str_val)  # for int, long and float
    except ValueError:
        try:
            complex(str_val)  # for complex
        except ValueError:
            return False
    return True


def zero_pad_num_str(
    num_val: float,
    str_len: int = 3,
    round_num: bool = False,
    round_n_digts: int = 0,
    integerise: bool = False,
    absolute: bool = False,
    gain: float = 1,
):
    """
    A function which zero pads a number to make a string

    :param num_val: number value to be processed.
    :param str_len: the number of characters in the output string.
    :param round_num: boolean whether to round the input number value.
    :param round_n_digts: If rounding, the number of digits following decimal
                          points to round to.
    :param integerise: boolean whether to integerise the input number
    :param absolute: make number positive.
    :param gain: apply a gain before integerising.
    :return: string with the padded numeric value.

    """
    if absolute:
        num_val = abs(num_val)
    if round_num:
        num_val = round(num_val, round_n_digts)
    if integerise:
        num_val = int(num_val * gain)

    num_str = "{}".format(num_val)
    num_str = num_str.zfill(str_len)
    return num_str


def powerset_iter(in_set: list):
    """
    A function which returns an iterator (generator) for all the subsets
    of the inputted set (i.e., the powerset)

    :params inset: the input set for which the powerset will be produced

    """
    if len(in_set) <= 1:
        yield in_set
        yield []
    else:
        for item in powerset_iter(in_set[1:]):
            yield [in_set[0]] + item
            yield item


def powerset_lst(in_set: list, min_items: int = 0):
    """
    A function which returns a list for all the subsets
    of the inputted set (i.e., the powerset)

    :params inset: the input set for which the powerset will be produced
    :params min_items: Optional parameter specifying the minimum number
                       of items in the output sets. If 0 or below then
                       ignored. Default is 0.

    """
    out_pset = []
    for subset in powerset_iter(in_set):
        if min_items > 0:
            if len(subset) >= min_items:
                out_pset.append(subset)
        else:
            out_pset.append(subset)
    return out_pset


def get_environment_variable(var: str):
    """
    A function to get an environmental variable, if variable is not present
    returns None.

    :param var: the name of the environmental variable to get the value for.
    :return: value of env var.

    """
    outVar = None
    try:
        outVar = os.environ[var]
    except Exception:
        outVar = None
    return outVar


def num_process_cores():
    """
    A functions which returns the number of processing cores available on the machine

    :return: int

    """
    import multiprocessing

    return multiprocessing.cpu_count()


def read_text_file_no_new_lines(input_file: str):
    """
    Read a text file into a single string removing new lines.

    :param input_file: File path to the input file.
    :return: string

    """
    txtStr = ""
    try:
        dataFile = open(input_file, "r")
        for line in dataFile:
            txtStr += line.strip()
        dataFile.close()
    except Exception as e:
        raise e
    return txtStr


def read_text_file_to_list(input_file: str):
    """
    Read a text file into a list where each line is an element in the list.

    :param input_file: File path to the input file.
    :return: list

    """
    outList = []
    try:
        dataFile = open(input_file, "r")
        for line in dataFile:
            line = line.strip()
            if line != "":
                outList.append(line)
        dataFile.close()
    except Exception as e:
        raise e
    return outList


def write_list_to_file(data_lst: list, out_file: str):
    """
    Write a list a text file, one line per item.

    :param data_lst: List of values to be written to the output file.
    :param out_file: File path to the output file.

    """
    try:
        f = open(out_file, "w")
        for item in data_lst:
            f.write(str(item) + "\n")
        f.flush()
        f.close()
    except Exception as e:
        raise e


def write_data_to_file(data_val, out_file: str):
    """
    Write some data (a string or can be converted to a string using str(data_val) to
    an output text file.

    :param data_val: Data to be written to the output file.
    :param out_file: File path to the output file.

    """
    try:
        f = open(out_file, "w")
        f.write(str(data_val) + "\n")
        f.flush()
        f.close()
    except Exception as e:
        raise e


def write_dict_to_json(data_dict: dict, out_file: str):
    """
    Write some data to a JSON file. The data would commonly be structured as a dict
    but could also be a list.

    :param data_dict: The dict (or list) to be written to the output JSON file.
    :param out_file: The file path to the output file.

    """
    with open(out_file, "w") as fp:
        json.dump(
            data_dict,
            fp,
            cls=RSGISNumpyArrayEncoder,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
            ensure_ascii=False,
        )


def write_dict_to_json_gz(data_dict: dict, out_file: str, encoding: str = "utf-8"):
    """
    Write some data to a JSON file compressed with gzip. The data would commonly be
    structured as a dict but could also be a list.

    :param data_dict: The dict (or list) to be written to the output JSON file.
    :param out_file: The file path to the output file.
    :param encoding: the character set encoding (default: utf-8)

    """
    import gzip

    json_str = json.dumps(
        data_dict,
        cls=RSGISNumpyArrayEncoder,
        sort_keys=True,
        indent=4,
        separators=(",", ": "),
        ensure_ascii=False,
    )
    json_bytes = json_str.encode(encoding)

    with gzip.GzipFile(out_file, "w") as fout:
        fout.write(json_bytes)


def read_json_to_dict(input_file: str):
    """
    Read a JSON file. Will return a list or dict.

    :param input_file: input JSON file path.

    """
    with open(input_file) as f:
        data = json.load(f)
    return data


def read_gz_json_to_dict(input_file: str, encoding: str = "utf-8"):
    """
    Read a JSON file. Will return a list or dict.

    :param input_file: input JSON file path.
    :param encoding: the character set encoding (default: utf-8)

    """
    import gzip

    with gzip.GzipFile(input_file, "r") as fin:
        json_bytes = fin.read()

    json_str = json_bytes.decode(encoding)
    data = json.loads(json_str)
    return data


def create_var_list(in_vals_lsts: dict, val_dict: dict = None):
    """
    A function which will produce a list of dictionaries with all the combinations
    of the input variables listed (i.e., the powerset).

    :param in_vals_lsts: dictionary with each value having a list of values.
    :param val_dict: variable used in iterative nature of function which lists
                     the variable for which are still to be looped through. Would
                     normally not be provided by the user as default is None. Be
                     careful if you set as otherwise you might not get a correct
                     or valid result.
    :returns: list of dictionaries with the same keys are the input but only a
              single value will be associate with key rather than a list.

    .. code:: python

        import rsgislib.tools.utils
        seg_vars_ranges = dict()
        seg_vars_ranges['k'] = [5, 10, 20, 30, 40, 50, 60, 80, 100, 120]
        seg_vars_ranges['d'] = [10, 20, 50, 100, 200, 1000, 10000]
        seg_vars_ranges['minsize'] = [5, 10, 20, 50, 100, 200]
        seg_vars = rsgislib.tools.utils.create_var_list(seg_vars_ranges)

    """
    out_vars = []
    if (in_vals_lsts is None) and (val_dict is not None):
        out_val_dict = dict()
        for key in val_dict.keys():
            out_val_dict[key] = val_dict[key]
        out_vars.append(out_val_dict)
    elif in_vals_lsts is not None:
        if len(in_vals_lsts.keys()) > 0:
            key = list(in_vals_lsts.keys())[0]
            vals_arr = in_vals_lsts[key]
            next_vals_lsts = dict()
            for ckey in in_vals_lsts.keys():
                if ckey != key:
                    next_vals_lsts[ckey] = in_vals_lsts[ckey]

            if len(next_vals_lsts.keys()) == 0:
                next_vals_lsts = None

            if val_dict is None:
                val_dict = dict()

            for val in vals_arr:
                c_val_dict = dict()
                for ckey in val_dict.keys():
                    c_val_dict[ckey] = val_dict[ckey]
                c_val_dict[key] = val
                c_out_vars = create_var_list(next_vals_lsts, c_val_dict)
                out_vars = out_vars + c_out_vars
    return out_vars


def in_bounds(x: numpy.array, lower: float, upper: float, upper_strict: bool = False):
    """
    Checks whether a value or array of values is within specified bounds.

    :param x: value or numpy array of values to check.
    :param lower: lower bound
    :param upper: upper bound
    :param upper_strict: True is less than upper; False is less than equal to upper
    :return: boolean

    """
    if upper_strict:
        return lower <= numpy.min(x) and numpy.max(x) < upper
    else:
        return lower <= numpy.min(x) and numpy.max(x) <= upper


def mixed_signs(x: numpy.array):
    """
    Check whether an array of numbers has a mix of positive and negative values.

    :param x: list of values.
    :return: boolean

    """
    return numpy.min(x) < 0 and numpy.max(x) >= 0


def negative(x: numpy.array):
    """
    Is the maximum number in the list negative.
    :param x: list of values

    :return: boolean

    """
    return numpy.max(x) < 0


def is_odd(number: float):
    """
    A function which tests whether a number is odd

    :param number: number value to test.
    :return: True = input number is odd; False = input number is even

    """
    if (number % 2) != 0:
        return True
    return False


def hex_to_rgb(hex_str: str):
    """
    A function which converts an hexadecimal colour representation to RGB values
    between 0 and 255.

    For example: #b432be is equal to: 180, 50, 190

    :param hex_str: Input hex string which can be either 7 or 6 characters long.
                    If 7 characters then the first character will be a #.
    :return: R, G, B tuple

    .. code:: python

        import rsgislib.tools.utils
        r, g, b = rsgislib.tools.utils.hex_to_rgb("#b432be")

    """
    if hex_str[0] == "#":
        hex_str = hex_str[1:]
    if len(hex_str) != 6:
        raise rsgislib.RSGISPyException(
            "String must be of length 6 or 7 if starting with #"
        )

    r_hex = hex_str[0:2]
    g_hex = hex_str[2:4]
    b_hex = hex_str[4:6]
    return int(r_hex, 16), int(g_hex, 16), int(b_hex, 16)


def remove_repeated_chars(str_val: str, repeat_char: str):
    """
    A function which removes repeated characters within a string for the
    specified character

    :param str_val: The input string.
    :param repeat_char: The character
    :return: string without repeat_char

    """
    if len(repeat_char) != 1:
        raise rsgislib.RSGISPyException("The repeat character has multiple characters.")
    out_str = ""
    p = ""
    for c in str_val:
        if c == repeat_char:
            if c != p:
                out_str += c
        else:
            out_str += c
        p = c
    return out_str


def check_str(
    str_val: str,
    rm_non_ascii: bool = False,
    rm_dashs: bool = False,
    rm_spaces: bool = False,
    rm_punc: bool = False,
):
    """
    A function which can check a string removing spaces (replaced with underscores),
    remove punctuation and any non ascii characters.

    :param str_val: the input string to be processed.
    :param rm_non_ascii: If True (default False) remove any non-ascii characters
                         from the string
    :param rm_dashs: If True (default False) remove any dashs from the string
                     and replace with underscores.
    :param rm_spaces: If True (default False) remove any spaces from the string.
    :param rm_punc: If True (default False) remove any punctuation
                    (other than '_' or '-') from the string.
    :return: returns a string outputted from the processing.

    """
    str_val_tmp = str_val.strip()

    if rm_non_ascii:
        str_val_tmp_ascii = ""
        for c in str_val_tmp:
            if (
                (c in string.ascii_letters)
                or (c in string.punctuation)
                or (c in string.digits)
                or (c == " ")
            ):
                str_val_tmp_ascii += c
        str_val_tmp = str_val_tmp_ascii

    if rm_dashs:
        str_val_tmp = str_val_tmp.replace("-", "_")
        str_val_tmp = remove_repeated_chars(str_val_tmp, "_")

    if rm_spaces:
        str_val_tmp = str_val_tmp.replace(" ", "_")
        str_val_tmp = remove_repeated_chars(str_val_tmp, "_")

    if rm_punc:
        for punct in string.punctuation:
            if (punct != "_") and (punct != "-"):
                str_val_tmp = str_val_tmp.replace(punct, "")
        str_val_tmp = remove_repeated_chars(str_val_tmp, "_")

    return str_val_tmp


def get_days_since(year: int, day_of_year: int, base_date: datetime.date):
    """
    Calculate the number of days from a base data to a defined year/day.

    :param year: int with year XXXX (e.g., 2020)
    :param day_of_year: int with the day within the year (1-365)
    :param base_date: a datetime.date object
    :return: int (n days)

    """
    if year < base_date.year:
        raise rsgislib.RSGISPyException("The year specified is before the base date.")
    date_val = datetime.date(year=int(year), month=1, day=1)
    date_val = date_val + datetime.timedelta(days=int(day_of_year - 1))
    return (date_val - base_date).days


def get_days_since_date(year: int, month: int, day, base_date: datetime.date):
    """
    Calculate the number of days from a base data to a defined year/day.

    :param year: int with year XXXX (e.g., 2020)
    :param month: int month in year (1-12) (e.g., 6)
    :param day: int with the day within the month (1-31) (e.g., 20)
    :param base_date: a datetime.date object
    :return: int (n days)

    """
    if year < base_date.year:
        raise rsgislib.RSGISPyException("The year specified is before the base date.")
    date_val = datetime.date(year=int(year), month=int(month), day=int(day))
    return (date_val - base_date).days


def dict_struct_does_path_exist(dict_struct_obj: dict, tree_sequence: list):
    """
    A function which tests whether a path exists within JSON file.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings specifying the path within the structure
    :return: boolean

    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    pathExists = True
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            pathExists = False
            break
    return pathExists


def dict_struct_get_str_value(
    dict_struct_obj: dict, tree_sequence: list, valid_values: list = None
):
    """
    A function which retrieves a single string value from a JSON structure.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings specifying the path within the structure
    :param valid_values: An optional list of valid values. An exception will be
                         thrown if a list is provided and the string value found
                         is not within the list.
    :return: the string value at the path provided.

    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            raise rsgislib.RSGISPyException("Could not find '{}'".format(steps_str))
    if valid_values is not None:
        if curr_dict_struct_obj not in valid_values:
            raise rsgislib.RSGISPyException(
                "'{}' is not within the list of valid values.".format(
                    curr_dict_struct_obj
                )
            )
    return curr_dict_struct_obj


def dict_struct_get_boolean_value(dict_struct_obj: dict, tree_sequence: list):
    """
    A function which retrieves a single boolean value from a JSON structure.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings specifying the path within the structure
    :return: boolean value.

    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            raise rsgislib.RSGISPyException("Could not find '{}'".format(steps_str))
    if type(curr_dict_struct_obj).__name__ == "bool":
        rtn_bool = curr_dict_struct_obj
    else:
        raise rsgislib.RSGISPyException(
            "'{}' is not 'True' or 'False'.".format(curr_dict_struct_obj)
        )
    return rtn_bool


def dict_struct_get_date_value(
    dict_struct_obj: dict, tree_sequence: list, date_format: str = "%Y-%m-%d"
):
    """
    A function which retrieves a single date value from a JSON structure.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings specifying the path within the structure
    :param date_format: a string or list of strings for the date/time format
                             to be parsed by datetime.datetime.strptime.
    :return: a date object with the value at the path specified.

    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            raise rsgislib.RSGISPyException("Could not find '{}'".format(steps_str))

    if type(date_format) is str:
        try:
            out_date_obj = datetime.datetime.strptime(
                curr_dict_struct_obj, date_format
            ).date()
        except Exception as e:
            raise rsgislib.RSGISPyException(e)
    elif type(date_format) is list:
        found = False
        except_obj = None
        for date_format_str in date_format:
            try:
                out_date_obj = datetime.datetime.strptime(
                    curr_dict_struct_obj, date_format_str
                ).date()
                found = True
                break
            except Exception as e:
                except_obj = e
        if not found:
            raise rsgislib.RSGISPyException(except_obj)
    else:
        raise rsgislib.RSGISPyException(
            "Do not know what the type is of date_format variable."
        )

    return out_date_obj


def dict_struct_get_datetime_value(
    dict_struct_obj: dict,
    tree_sequence: list,
    date_time_format: str = "%Y-%m-%dT%H:%M:%S.%f",
):
    """
    A function which retrieves a single date value from a JSON structure.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings specifying the path within the structure
    :param date_time_format: a string or list of strings for the date/time format
                             to be parsed by datetime.datetime.strptime.
    :return: datetime object with the value at the path provided.
    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            raise rsgislib.RSGISPyException("Could not find '{}'".format(steps_str))

    curr_dict_struct_obj = curr_dict_struct_obj.replace("Z", "")
    if type(date_time_format) is str:
        try:
            out_datetime_obj = datetime.datetime.strptime(
                curr_dict_struct_obj, date_time_format
            )
        except Exception as e:
            raise rsgislib.RSGISPyException(e)
    elif type(date_time_format) is list:
        found = False
        except_obj = None
        for date_time_format_str in date_time_format:
            try:
                out_datetime_obj = datetime.datetime.strptime(
                    curr_dict_struct_obj, date_time_format_str
                )
                found = True
                break
            except Exception as e:
                except_obj = e
        if not found:
            raise rsgislib.RSGISPyException(except_obj)
    else:
        raise rsgislib.RSGISPyException(
            "Do not know what the type is of date_time_format variable."
        )

    return out_datetime_obj


def dict_struct_get_str_list_value(
    dict_struct_obj: dict, tree_sequence: list, valid_values: list = None
):
    """
    A function which retrieves a list of string values from a JSON structure.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings specifying the path within the structure
    :param valid_values: An optional list of valid values. An exception will be
                         thrown if a list is provided and a string value found
                         is not within the list.
    :return: list of strings

    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            raise rsgislib.RSGISPyException("Could not find '{}'".format(steps_str))

    if type(curr_dict_struct_obj).__name__ != "list":
        raise rsgislib.RSGISPyException("Retrieved value is not a list.")
    if valid_values is not None:
        for val in curr_dict_struct_obj:
            if type(val).__name__ != "str":
                raise rsgislib.RSGISPyException(
                    "'{}' is not of type string.".format(val)
                )
            if val not in valid_values:
                raise rsgislib.RSGISPyException(
                    "'{}' is not within the list of valid values.".format(val)
                )
    return curr_dict_struct_obj


def dict_struct_get_numeric_value(
    dict_struct_obj: dict,
    tree_sequence: list,
    valid_lower: float = None,
    valid_upper: float = None,
):
    """
    A function which retrieves a single numeric value from a JSON structure.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings specifying the path within the structure
    :param valid_lower: optional lower valid bounds, error throw if outside the bounds.
    :param valid_upper: optional upper valid bounds, error throw if outside the bounds.
    :return: numeric value at path

    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            raise rsgislib.RSGISPyException("Could not find '{}'".format(steps_str))

    out_value = 0.0
    if (type(curr_dict_struct_obj).__name__ == "int") or (
        type(curr_dict_struct_obj).__name__ == "float"
    ):
        out_value = curr_dict_struct_obj
    elif type(curr_dict_struct_obj).__name__ == "str":
        if curr_dict_struct_obj.isnumeric():
            out_value = float(curr_dict_struct_obj)
        else:
            try:
                out_value = float(curr_dict_struct_obj)
            except:
                raise rsgislib.RSGISPyException(
                    "The identified value is not numeric '{}'".format(steps_str)
                )
    else:
        raise rsgislib.RSGISPyException(
            "The identified value is not numeric '{}'".format(steps_str)
        )

    if valid_lower is not None:
        if out_value < valid_lower:
            raise rsgislib.RSGISPyException(
                "'{}' is less than the defined valid range.".format(out_value)
            )
    if valid_upper is not None:
        if out_value > valid_upper:
            raise rsgislib.RSGISPyException(
                "'{}' is higher than the defined valid range.".format(out_value)
            )
    return out_value


def dict_struct_get_list_value(dict_struct_obj: dict, tree_sequence: list):
    """
    A function which retrieves a list of values from a JSON structure.

    :param dict_struct_obj: the structure of dicts (i.e., JSON) to be searched
    :param tree_sequence: list of strings  specifying the path within the structure
    :return: list of values

    """
    curr_dict_struct_obj = dict_struct_obj
    steps_str = ""
    for tree_step in tree_sequence:
        steps_str = steps_str + ":" + tree_step
        if tree_step in curr_dict_struct_obj:
            curr_dict_struct_obj = curr_dict_struct_obj[tree_step]
        else:
            raise rsgislib.RSGISPyException("Could not find '{}'".format(steps_str))

    if type(curr_dict_struct_obj).__name__ != "list":
        raise rsgislib.RSGISPyException("Retrieved value is not a list.")
    return curr_dict_struct_obj



import os
import pytest

SKLEARN_NOT_AVAIL = False
try:
    import sklearn
except ImportError:
    SKLEARN_NOT_AVAIL = True

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
REGRESS_DATA_DIR = os.path.join(DATA_DIR, "regression")


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_get_obj_params():
    import rsgislib.regression.regresssklearn

    rsgislib.regression.regresssklearn.get_ann_obj_params(1)
    rsgislib.regression.regresssklearn.get_en_obj_params(1)
    rsgislib.regression.regresssklearn.get_knn_obj_params(1)
    rsgislib.regression.regresssklearn.get_kr_obj_params(1)
    rsgislib.regression.regresssklearn.get_et_obj_params(1)
    rsgislib.regression.regresssklearn.get_pls_obj_params(1)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_create_search_obj():
    import rsgislib.regression.regresssklearn
    import numpy
    from sklearn.ensemble import ExtraTreesRegressor

    et_grid = {
        "n_estimators": numpy.arange(100, 200, 100, dtype="uint16").tolist(),
        "max_features": numpy.arange(1, 2, 1, dtype="uint8").tolist(),
        "max_depth": numpy.arange(1, 2, 1, dtype="uint8").tolist(),
    }
    et_obj = ExtraTreesRegressor()
    rsgislib.regression.regresssklearn.create_search_obj(
        et_obj, et_grid, n_runs=25, n_cv=2, n_cores=1
    )


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_perform_search_param_opt(tmp_path):
    import rsgislib.regression.regresssklearn
    import rsgislib.vectorattrs
    import numpy
    from sklearn.ensemble import ExtraTreesRegressor

    et_grid = {
        "n_estimators": numpy.arange(100, 200, 100, dtype="uint16").tolist(),
        "max_features": numpy.arange(1, 2, 1, dtype="uint8").tolist(),
        "max_depth": numpy.arange(1, 2, 1, dtype="uint8").tolist(),
    }
    et_obj = ExtraTreesRegressor()
    skl_srch_obj = rsgislib.regression.regresssklearn.create_search_obj(
        et_obj, et_grid, n_runs=25, n_cv=2, n_cores=1
    )

    vec_file = os.path.join(REGRESS_DATA_DIR, "sample_pts_test.geojson")
    vec_lyr = "sample_pts_test"

    y = rsgislib.vectorattrs.get_vec_cols_as_array(vec_file, vec_lyr, cols=["value"])
    x = rsgislib.vectorattrs.get_vec_cols_as_array(
        vec_file, vec_lyr, cols=["re_b4", "re_b5", "re_b6", "nir_b7", "nir_b8"]
    )

    opt_params_file = os.path.join(tmp_path, "out.json")
    rsgislib.regression.regresssklearn.perform_search_param_opt(
        opt_params_file, x, y, skl_srch_obj, data_scaler=None
    )

    assert os.path.exists(opt_params_file)


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_perform_kfold_fit():
    import rsgislib.regression.regresssklearn
    import rsgislib.vectorattrs
    from sklearn.ensemble import ExtraTreesRegressor

    et_obj = ExtraTreesRegressor()

    vec_file = os.path.join(REGRESS_DATA_DIR, "sample_pts_train.geojson")
    vec_lyr = "sample_pts_train"

    y = rsgislib.vectorattrs.get_vec_cols_as_array(vec_file, vec_lyr, cols=["value"])
    x = rsgislib.vectorattrs.get_vec_cols_as_array(
        vec_file, vec_lyr, cols=["re_b4", "re_b5", "re_b6", "nir_b7", "nir_b8"]
    )
    rsgislib.regression.regresssklearn.perform_kfold_fit(
        et_obj, x, y, n_splits=2, repeats=2, shuffle=True, data_scaler=None
    )


@pytest.mark.skipif(SKLEARN_NOT_AVAIL, reason="scikit-learn dependency not available")
def test_apply_regress_sklearn_mdl(tmp_path):
    import rsgislib.regression.regresssklearn
    import rsgislib.vectorattrs
    from sklearn.ensemble import ExtraTreesRegressor

    et_obj = ExtraTreesRegressor()

    vec_file = os.path.join(REGRESS_DATA_DIR, "sample_pts_train.geojson")
    vec_lyr = "sample_pts_train"

    y = rsgislib.vectorattrs.get_vec_cols_as_array(vec_file, vec_lyr, cols=["value"])
    x = rsgislib.vectorattrs.get_vec_cols_as_array(
        vec_file, vec_lyr, cols=["re_b4", "re_b5", "re_b6", "nir_b7", "nir_b8"]
    )

    et_obj.fit(x, y)

    s2_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset.kea")
    vld_msk_img = os.path.join(DATA_DIR, "sen2_20210527_aber_subset_vldmsk.kea")
    out_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.regression.regresssklearn.apply_regress_sklearn_mdl(
        et_obj,
        1,
        s2_img,
        [4, 5, 6, 7, 8],
        vld_msk_img,
        1,
        out_img,
        gdalformat="KEA",
        out_band_names=None,
        calc_stats=True,
        out_no_date_val=0.0,
    )

    assert os.path.exists(out_img)



import os

DATA_DIR = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "data", "imagemorphology"
)


def test_create_circular_op(tmp_path):
    import rsgislib.imagemorphology

    out_file = os.path.join(tmp_path, "CircularOp.gmtxt")
    op_size = 5
    rsgislib.imagemorphology.create_circular_op(out_file, op_size)

    assert os.path.exists(out_file)


def test_image_dilate(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgDilate_OpFile.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_dilate(
        input_img, output_img, morph_op_file, use_op_file, op_size, gdalformat, datatype
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgDilate_OpFile.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_dilate_combined_out(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgDilate_combined.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_dilate_combined_out(
        input_img, output_img, morph_op_file, use_op_file, op_size, gdalformat, datatype
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgDilate_combined.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_erode(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgErode.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_erode(
        input_img, output_img, morph_op_file, use_op_file, op_size, gdalformat, datatype
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgErode.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_erode_combined_out(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgErode_combined.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_erode_combined_out(
        input_img, output_img, morph_op_file, use_op_file, op_size, gdalformat, datatype
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgErode_combined.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_gradiant(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgGradiant.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_32FLOAT
    rsgislib.imagemorphology.image_gradiant(
        input_img, output_img, morph_op_file, use_op_file, op_size, gdalformat, datatype
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgGradiant.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_gradiant_combined_out(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgGradiant_combined.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_32FLOAT
    rsgislib.imagemorphology.image_gradiant_combined_out(
        input_img, output_img, morph_op_file, use_op_file, op_size, gdalformat, datatype
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgGradiant_combined.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_local_minima(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgLocalMinima.kea")
    sequencial_out = True
    allow_equal = True
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_local_minima(
        input_img,
        output_img,
        sequencial_out,
        allow_equal,
        morph_op_file,
        use_op_file,
        op_size,
        gdalformat,
        datatype,
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgLocalMinima.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_local_minima_combined_out(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgLocalMinimaCombined.kea")
    sequencial_out = True
    allow_equal = True
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_local_minima_combined_out(
        input_img,
        output_img,
        sequencial_out,
        allow_equal,
        morph_op_file,
        use_op_file,
        op_size,
        gdalformat,
        datatype,
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgLocalMinimaCombined.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    assert img_eq


def test_image_opening(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgOpening.kea")
    tmp_img = os.path.join(tmp_path, "sen2_20210527_aber_imgOpening_tmp.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    niters = 1
    rsgislib.imagemorphology.image_opening(
        input_img,
        output_img,
        tmp_img,
        morph_op_file,
        use_op_file,
        op_size,
        gdalformat,
        datatype,
        niters,
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgOpening.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    print(prop_match)
    assert img_eq


def test_image_closing(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgClosing.kea")
    tmp_img = os.path.join(tmp_path, "sen2_20210527_aber_imgClosing_tmp.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    niters = 1
    rsgislib.imagemorphology.image_closing(
        input_img,
        output_img,
        tmp_img,
        morph_op_file,
        use_op_file,
        op_size,
        gdalformat,
        datatype,
        niters,
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgClosing.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    print(prop_match)
    assert img_eq


def test_image_black_top_hat(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgBlackTopHat.kea")
    tmp_img = os.path.join(tmp_path, "sen2_20210527_aber_imgBlackTopHat_tmp.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_black_top_hat(
        input_img,
        output_img,
        tmp_img,
        morph_op_file,
        use_op_file,
        op_size,
        gdalformat,
        datatype,
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgBlackTopHat.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    print(prop_match)
    assert img_eq


def test_image_white_top_hat(tmp_path):
    import rsgislib.imagemorphology
    import rsgislib.imagecalc

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber_NDVI_lt5_bin.kea")
    output_img = os.path.join(tmp_path, "sen2_20210527_aber_imgWhiteTopHat.kea")
    tmp_img = os.path.join(tmp_path, "sen2_20210527_aber_imgWhiteTopHat_tmp.kea")
    morph_op_file = os.path.join(DATA_DIR, "CircularOp.gmtxt")
    use_op_file = True
    op_size = 5
    gdalformat = "KEA"
    datatype = rsgislib.TYPE_8UINT
    rsgislib.imagemorphology.image_white_top_hat(
        input_img,
        output_img,
        tmp_img,
        morph_op_file,
        use_op_file,
        op_size,
        gdalformat,
        datatype,
    )

    ref_img = os.path.join(DATA_DIR, "sen2_20210527_aber_imgWhiteTopHat.kea")
    img_eq, prop_match = rsgislib.imagecalc.are_imgs_equal(ref_img, output_img)
    print(prop_match)
    assert img_eq



import os
import pytest

DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")

GEOPANDAS_NOT_AVAIL = False
try:
    import geopandas
except ImportError:
    GEOPANDAS_NOT_AVAIL = True


def test_rasterise_vec_lyr(tmp_path):
    import rsgislib.vectorutils.createrasters

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
        vec_file, vec_lyr, input_img, output_img, gdalformat="KEA"
    )
    assert os.path.exists(output_img)


def test_rasterise_vec_lyr_burn_col(tmp_path):
    import rsgislib.vectorutils.createrasters

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.vectorutils.createrasters.rasterise_vec_lyr(
        vec_file, vec_lyr, input_img, output_img, gdalformat="KEA", att_column="val"
    )
    assert os.path.exists(output_img)


def test_rasterise_vec_lyr_obj(tmp_path):
    import rsgislib.imageutils
    import rsgislib.vectorutils
    import rsgislib.vectorutils.createrasters

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_obj_ds, vec_lyr_obj = rsgislib.vectorutils.open_gdal_vec_lyr(vec_file, vec_lyr)

    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, "KEA", rsgislib.TYPE_16UINT
    )

    rsgislib.vectorutils.createrasters.rasterise_vec_lyr_obj(vec_lyr_obj, output_img)
    vec_obj_ds = None
    assert os.path.exists(output_img)


def test_rasterise_vec_lyr_obj_burn_col(tmp_path):
    import rsgislib.imageutils
    import rsgislib.vectorutils
    import rsgislib.vectorutils.createrasters

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    vec_obj_ds, vec_lyr_obj = rsgislib.vectorutils.open_gdal_vec_lyr(vec_file, vec_lyr)

    output_img = os.path.join(tmp_path, "out_img.kea")
    rsgislib.imageutils.create_copy_img(
        input_img, output_img, 1, 0, "KEA", rsgislib.TYPE_16UINT
    )

    rsgislib.vectorutils.createrasters.rasterise_vec_lyr_obj(
        vec_lyr_obj, output_img, att_column="val"
    )
    vec_obj_ds = None
    assert os.path.exists(output_img)


def test_copy_vec_to_rat_exception(tmp_path):
    import rsgislib
    import rsgislib.vectorutils.createrasters

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    output_img = os.path.join(tmp_path, "out_img.kea")
    with pytest.raises(rsgislib.RSGISPyException):
        rsgislib.vectorutils.createrasters.copy_vec_to_rat(
            vec_file, vec_lyr, input_img, output_img
        )


@pytest.mark.skipif(GEOPANDAS_NOT_AVAIL, reason="geopandas dependency not available")
def test_copy_vec_to_rat(tmp_path):
    import rsgislib.vectorutils.createrasters
    import rsgislib.vectorattrs

    input_img = os.path.join(DATA_DIR, "sen2_20210527_aber.kea")
    vec_file = os.path.join(DATA_DIR, "aber_osgb_multi_polys.geojson")
    vec_lyr = "aber_osgb_multi_polys"

    out_vec_file = os.path.join(tmp_path, "aber_osgb_multi_polys_fid.geojson")
    out_vec_lyr = "aber_osgb_multi_polys_fid"

    rsgislib.vectorattrs.add_fid_col(
        vec_file,
        vec_lyr,
        out_vec_file,
        out_vec_lyr,
        out_format="GEOJSON",
        out_col="FID",
    )

    output_img = os.path.join(tmp_path, "out_img.kea")

    rsgislib.vectorutils.createrasters.copy_vec_to_rat(
        out_vec_file, out_vec_lyr, input_img, output_img, fid_col="FID"
    )

    assert os.path.exists(output_img)



#!/usr/bin/env python
"""
The tools.projection module contains some useful tools for working with projections.
"""

import numpy
from osgeo import osr


def get_epsg_code_from_wkt(wkt_str: str) -> int:
    """
    Using GDAL to return the EPSG code for inputted WKT string.

    :return: the EPSG code.

    """
    epsg_code = None
    try:
        spatRef = osr.SpatialReference()
        spatRef.ImportFromWkt(wkt_str)
        spatRef.AutoIdentifyEPSG()
        epsg_code = spatRef.GetAuthorityCode(None)
        epsg_code = int(epsg_code)
    except Exception:
        epsg_code = None
    return epsg_code


def get_wkt_from_epsg_code(epsg_code: int) -> str:
    """
    Using GDAL to return the WKT string for inputted EPSG Code.

    :param epsg_code: integer variable of the epsg code.

    :return: string with WKT representation of the projection.

    """
    wkt_str = None
    try:
        spatRef = osr.SpatialReference()
        spatRef.ImportFromEPSG(epsg_code)
        wkt_str = spatRef.ExportToWkt()
    except Exception:
        wkt_str = None
    return wkt_str


def get_osr_prj_obj(epsg_code: int) -> osr.SpatialReference():
    """
    A function which returns an OSR SpatialReference object
    for a given EPSG code.

    :param epsg_code: An EPSG code for the projection. Must be an integer.
    :return: osr.SpatialReference object.

    """
    spat_ref = osr.SpatialReference()
    spat_ref.ImportFromEPSG(int(epsg_code))
    return spat_ref


def degrees_to_metres(
    latitude: float, lon_size: float, lat_size: float
) -> (float, float):
    """
    Convert pixel sizes or distances in degrees to metres.

    :param latitude: latitude
    :param lon_size: numpy array of x pixel sizes (degrees)
    :param lat_size: numpy array of y pixel sizes (degrees)
    :return: x_size (numpy array of x pixel sizes (m)), y_size (numpy array of y pixel sizes (m))

    .. code:: python

        from rsgislib import tools
        x_size, y_size = tools.degrees_to_metres(52,1.0,1.0)

    """

    # Set up parameters for ellipse
    # Semi-major and semi-minor for WGS-84 ellipse
    ellipse = [6378137.0, 6356752.314245]

    radlat = numpy.deg2rad(latitude)

    Rsq = (ellipse[0] * numpy.cos(radlat)) ** 2 + (ellipse[1] * numpy.sin(radlat)) ** 2
    Mlat = (ellipse[0] * ellipse[1]) ** 2 / (Rsq**1.5)
    Nlon = ellipse[0] ** 2 / numpy.sqrt(Rsq)
    x_size = numpy.pi / 180 * numpy.cos(radlat) * Nlon * lon_size
    y_size = numpy.pi / 180 * Mlat * lat_size

    return x_size, y_size


def metres_to_degrees(latitude: float, x_size: float, y_size: float) -> (float, float):
    """
    Convert pixel sizes or distances in metres to degrees.

    :param latitude: latitude
    :param x_size: numpy array of x pixel sizes (m)
    :param y_size: numpy array of y pixel sizes (m)
    :return: lon_size (numpy array of x pixel sizes (degrees)), lat_size (numpy array of y pixel sizes (degrees))

    .. code:: python

        from rsgislib import tools
        x_size, y_size = tools.metres_to_degrees(52, 1.0, 1.0)

    """

    # Set up parameters for ellipse
    # Semi-major and semi-minor for WGS-84 ellipse
    ellipse = [6378137.0, 6356752.314245]

    radlat = numpy.deg2rad(latitude)

    Rsq = (ellipse[0] * numpy.cos(radlat)) ** 2 + (ellipse[1] * numpy.sin(radlat)) ** 2
    Mlat = (ellipse[0] * ellipse[1]) ** 2 / (Rsq**1.5)
    Nlon = ellipse[0] ** 2 / numpy.sqrt(Rsq)
    lon_size = x_size / (numpy.pi / 180 * numpy.cos(radlat) * Nlon)
    lat_size = y_size / (numpy.pi / 180 * Mlat)

    return lon_size, lat_size


def get_deg_coord_as_str(lat: float, lon: float, n_chars: int = 4) -> str:
    """

    :param lat:
    :param lon:
    :param n_chars:
    :return:
    """
    import rsgislib.tools.utils

    east_west = "e"
    if lon < 0:
        east_west = "w"
    north_south = "n"
    if lat < 0:
        north_south = "s"

    lat_str = rsgislib.tools.utils.zero_pad_num_str(
        lat,
        str_len=n_chars,
        round_num=True,
        round_n_digts=1,
        integerise=True,
        absolute=True,
        gain=10,
    )
    lon_str = rsgislib.tools.utils.zero_pad_num_str(
        lon,
        str_len=n_chars,
        round_num=True,
        round_n_digts=1,
        integerise=True,
        absolute=True,
        gain=10,
    )

    pos = "{}{}{}{}".format(north_south, lat_str, east_west, lon_str)
    return pos
